package simulation;

// IO.java, v2.3

/* -----------------------------------------------------------------------------
 * This package is based on Brinch Hansen's "Basic" classes.
 * Since some substantially changes were made, the copyright statement
 * and instructions below should be read with care.
 * -----------------------------------------------------------------------------

                   THE JAVA TEXT PROGRAM
                       28 April 1999
            Copyright (c) 1999 Per Brinch Hansen

This program defines a set of Java classes for text
processing. These text classes enable Java programs to
output text to screen and disk files and input text from
keyboard and disk files. The present version runs on
Macintosh, Unix, and Windows 95 systems.

Please follow these instructions to install the text classes
on a Unix system:

  1. Create a directory for your Java programs;

  2. Move this program text to your program directory as a
     file, named text.java;

  3. If you received the program text by email, use an
     editor to delete everything above the first line and
     everything below the last line of the program;

  4. Compile the text program into executable Java code. The
     compilation adds 11 code files to your program
     directory:

       Basic.class     Infile.class   Input.class
       Keyboard.class  Outfile.class  Output.class
       Random.class    Reader.class   Screen.class
       test.class      Writer.class

  5. Run a simple test of the text program by executing the
     compiled class, named test, as a self-contained Java
     program. If the compilation was successful, you will be
     asked to type your name, followed by return. When you
     have typed your name, say, Jane Doe, the program
     responds with the message: Welcome to Java, Jane Doe!

The Basic class defines Unix as the default system:

   public static final int OS = Unix;

If you wish to compile the text classes on a Macintosh (or
Windows 95) system, you must first use an editor to replace
the word, Unix, by the word, Mac (or OS_WINDOWS).

Please leave the text file, text.java, and the compiled
classes in your program directory. Java programs that use
the text classes must be compiled and run in the same
directory.

The following example shows a user program, named typist,
that uses the text classes:

    class typist extends Basic
    {
        public static void main(String param[]) throws Exception
        {
            Input in = new Input();
            Output out = new Output("results");
            while (in.more()) {
                out.write(in.read());
            }
            in.close();
            out.close();
        }
    }

The program extends the Basic class. The program copies text
from the keyboard to a new disk file, named results. The
input ends when you type a line consisting of the word, eof,
followed by return.
------------------------------------------------------------
This is the end of Brich Hansen's text.

Modifications by
    H.Moessenboeck, August 1999
    Theodorich Kopetzky, October 2000
    Ralf Hauber, March 2001

- Class Basic exports static variables in and out that
  represent the standard input and output stream.
  Methods of class Input and output don't throw exceptions
  any more. Thus a hello world program becomes as simple as

 	 class Hello extends Basic {
  		public static void main (String[] arg) {
  			out.writeLn("hello world");
  		}
 	 }

 - All methods of class Input (basically all read methods)
   set a boolean variable done,
   which indicates if the operation was successful. For example,
   reading a sequence of numbers becomes:

      x = readInt();
      while (IODone) {
      	...
      	x = readInt();
      }

- The constructor as well as the close method of class output
  set a boolean variable done indicating the success of the
  operation.

- Input and output files are always allocated and searched in
  the directory that contains the class files.

- Deprecated method DataInputStream.readLine avoided.

- Class Random has a new method to set the seed, which is now
  an instance variable instead of a static variable.

- major renaming to comply with our coding conventions

- Basic got convenience reads and writes

[v2.1 3/2001]

- Basic.java now has a version number.

[v2.2]

- Basic.java now is IO.java.

- renamed package: basicio.

- IODone is no longer public, instead there is an access method IO.isOk().

 [v2.3]

- renamed OS_WIN95 to OS_WINDOWS
- renamed SP to SPACE

------------------------------------------------------------*/

import java.io.*;
import java.lang.Math;

/* THE IO CLASS
=============================================================
Defines common constants and methods.
*/

public class IO
{
    public static final int
        OS_MAC = 0,
        OS_UNIX = 1,
        OS_WINDOWS = 2,
        OS = OS_WINDOWS; // set your operating system here ---------------------
    public static final String EOF_STRING = "eof";
    public static final char
        CR = '\r',
        EOF = '\uFFFF',
        NL = '\n',
        SPACE = ' ';

    private static Output out;
    private static Input in;

    private static boolean IODone;

    static /*initializer*/
    {
        out = null;
        in = null;
        try {
            out = new Output();
            in = new Input();
            IODone = true;
        } catch ( Exception e) {
            IODone = false;
            error( "could not open standard input/output");
        }
    }

    /** a simple assumption. Won't halt the program
    @param condition if condition is false a message will be printed
    */
    public static void assume( boolean condition)
    {
        assume( condition, "invalid assumption");
    }

    /** a simple assumption. Won't halt the program
    @param condition if condition is false msg will be printed
    @param msg the message to be printed
    */
    public static void assume( boolean condition, String msg)
    {
        if( !condition) {
            error( msg);
        }
    }

    /** Write an error message.
    @param msg the message to be printed
    */
    public static void error( String msg)
    {
        if( out != null) {
            out.writeLn( "*** error: " + msg);
        } else {
            System.err.println( "*** error: " + msg);
        }
    }

    /** Sets a new input encoding with enc.
    @param enc a String describing the input encoding e.g. "ISO8859_1"
    */
    public static void useInputEncoding( String enc) {
        try {
            in = new Input( new Keyboard ( enc));
        } catch( Exception e) {
        }
    }

    // --------------------------------------------------
    // reading simplicfications

    /**
    Read a character
    */
    public static char read() {
        char ch = in.read();
        IODone = in.done;
        return ch;
    }

    public static void readBlanks() {
        in.readBlanks();
        IODone = in.done;
    }

    public static boolean readBoolean() {
        boolean b = in.readBoolean();
        IODone = in.done;
        return b;
    }

    public static double readDouble() {
        double d = in.readDouble();
        IODone = in.done;
        return d;
    }

    public static int readInt() {
        int i = in.readInt();
        IODone = in.done;
        return i;
    }

    public static String readLine() {
        String s = in.readLine();
        IODone = in.done;
        return s;
    }

    public static void readLn() {
        in.readLn();
        IODone = in.done;
    }

    public static String readName() {
        String s = in.readName();
        IODone = in.done;
        return s;
    }

    public static void readNext() {
        in.readNext();
        IODone = in.done;
    }

    public static String readWord() {
        String s = in.readWord();
        IODone = in.done;
        return s;
    }

    // --------------------------------------------------
    // writing simplicfications

    public static void write(boolean value) {
        out.write(value);
    }

    public static void write(boolean value, int width) {
        out.write(value, width);
    }

    public static void write(char value) {
        out.write(value);
    }

    public static void write(char value, int width) {
        out.write(value, width);
    }

    public static void write(double value) {
        out.write(value);
    }

    public static void write(double value, int width) {
        out.write(value, width);
    }

    public static void write(int value) {
        out.write(value);
    }

    public static void write(int value, int width) {
        out.write(value, width);
    }

    public static void write(String value) {
        out.write(value);
    }

    public static void write(String value, int width) {
        out.write(value, width);
    }

    public static void writeLn() {
        out.writeLn();
    }

    public static void writeLn(boolean value) {
        out.writeLn(value);
    }

    public static void writeLn(boolean value, int width) {
        out.writeLn(value, width);
    }

    public static void writeLn(char value) {
        out.writeLn(value);
    }

    public static void writeLn(char value, int width) {
        out.writeLn(value, width);
    }

    public static void writeLn(double value) {
        out.writeLn(value);
    }

    public static void writeLn(double value, int width) {
        out.writeLn(value, width);
    }

    public static void writeLn(int value) {
        out.writeLn(value);
    }

    public static void writeLn(int value, int width) {
        out.writeLn(value, width);
    }

    public static void writeLn(String value) {
        out.writeLn(value);
    }

    public static void writeLn(String value, int width) {
        out.writeLn(value, width);
    }

    public static void writeSame(char value, int number) {
        out.writeSame(value, number);
    }

    public static void writeSpace(int number) {
        out.writeSpace(number);
    }

    public static boolean isOk() {
        return IODone;
    }

}

/* THE WRITER INTERFACE
=========================================================
Defines a writer as any class that includes a write
method for outputting a single character and a close
method for terminating output.
 */

interface Writer
{
    public void close() throws Exception;
    public void write(char value);
}

/* THE SCREEN CLASS
==========================================================
Implements a writer interface for outputting single
characters only to a screen.
 */

class Screen implements Writer {

    OutputStreamWriter out;

    public Screen() {
        FileOutputStream common = new FileOutputStream(FileDescriptor.out);
        try {
            out = new OutputStreamWriter(common, "ISO-8859-1");
        } catch (java.io.UnsupportedEncodingException e) {
            System.err.println ("*** ISO-8859-1 encoding unavailable ****");
            out = new OutputStreamWriter(common);
        }

    }

    public void close() {
        try {
            out.close ();
        } catch (java.io.IOException e) {
        }
    }

    public void write(char value) {
        try {
            out.write (value); out.flush();
            FileDescriptor.out.sync ();
        } catch (java.io.IOException e) {
            ;
        }
    }
}

/* THE OUTFILE CLASS
==========================================================
Implements a Writer interface for outputting single
characters only to a disk file. (Uses the class,
PrintStream, that has been deprecated, but is still
available in JDK 1.1.5.)
 */

class Outfile implements Writer
{
    private PrintStream out;

    public Outfile(String name) throws Exception
    {
        out = new PrintStream(new FileOutputStream(name));
    }

    public void close() throws Exception
    {
        if (out != null) {
            out.close(); out = null;
    } }

    public void write(char value)
    {
        if (value == IO.NL) {
            switch (IO.OS) {
                case IO.OS_MAC:
                    out.print(IO.CR);
                    break;

                case IO.OS_UNIX:
                    out.print(IO.NL);
                    break;

                case IO.OS_WINDOWS:
                    out.print(IO.CR);
                    out.print(IO.NL);
                    break;
            }
        } else {
            out.print(value);
        }
    }
}

/* THE OUTPUT CLASS
==========================================================
Used to output a text file to a screen or disk.
 */

class Output
{
    private Writer out;

    public boolean done;

    public Output()
    {
        out = new Screen();
        done = true;
    }

    public Output(String name)
    {
        try {
            out = new Outfile(name);
            done = true;
        } catch (Exception e) {
            IO.error("could not open output file " + name);
            done = false;
        }
    }

    public void close()
    {
        try {
            if (out != null) {
                out.close(); out = null;
            }
            done = true;
        } catch (Exception e) {
            IO.error("could not close output file");
            done = false;
        }
    }

    public void write(boolean value)
    {
        write(value, 1);
    }

    public void write(boolean value, int width)
    {
        String word = (String)(value?"true":"false");
        writeSpace(width - word.length());
        write(word);
    }

    public void write(char value)
    {
        out.write(value);
    }

    public void write(char value, int width)
    {
        write(value); writeSpace(width - 1);
    }

    public void write(double value)
    {
        write(value, 1);
    }

    public void write(double value, int width)
    {
        String numeral = String.valueOf(value);
        writeSpace(width - numeral.length());
        write(numeral);
    }

    public void write(int value)
    {
        write(value, 1);
    }

    public void write(int value, int width)
    {
        String numeral = String.valueOf((int)value);
        writeSpace(width - numeral.length());
        write(numeral);
    }

    public void write(String value)
    {
        write(value, 1);
    }

    public void write(String value, int width)
    {
        int length = value.length();
        for (int i = 0; i < length; i++)
        write(value.charAt(i));
        writeSpace(width - length);
    }

    public void writeLn()
    {
        write(IO.NL);
    }

    public void writeLn(boolean value)
    {
        writeLn(value, 1);
    }

    public void writeLn(boolean value, int width)
    {
        write(value, width); writeLn();
    }

    public void writeLn(char value)
    {
        writeLn(value, 1);
    }

    public void writeLn(char value, int width)
    {
        write(value, width); writeLn();
    }

    public void writeLn(double value)
    {
        writeLn(value, 1);
    }

    public void writeLn(double value, int width)
    {
        write(value, width); writeLn();
    }

    public void writeLn(int value)
    {
        writeLn(value, 1);
    }

    public void writeLn(int value, int width)
    {
        write(value, width); writeLn();
    }

    public void writeLn(String value)
    {
        writeLn(value, 1);
    }

    public void writeLn(String value, int width)
    {
        write(value, width); writeLn();
    }

    public void writeSame(char value, int number)
    {
        for (int i = 1; i <= number; i++) write(value);
    }

    public void writeSpace(int number)
    {
        writeSame(IO.SPACE, number);
    }
}

/* THE READER INTERFACE
===========================================================
Defines a reader as any class that includes a read
method for inputting a single character and a close
method for terminating input.
 */

interface Reader
{
    public void close() throws Exception;
    public char read();
}

/* THE Keyboard CLASS
===========================================================
Implements a Reader interface for inputting single
characters only from a Keyboard. (Uses the String
method, readLine, that has been deprecated, but is
still available in JDK 1.1.5.)
 */

class Keyboard implements Reader
{
    private InputStreamReader in;
    private char[] buffer = new char[512];
    private int typed, used;

    public Keyboard() throws Exception
    {
        this (null);
    }

    public Keyboard (String enc) throws Exception {
        FileInputStream common = new FileInputStream(FileDescriptor.in);

        if (enc == null) {
            in = new InputStreamReader(common);
        } else {
            try {
                in = new InputStreamReader(common, enc);
            } catch (java.io.UnsupportedEncodingException e) {
                System.err.println ("*** " + enc + " encoding unavailable ****");
                in = new InputStreamReader(common);
            }
        }
        typed = 0;
        used = 0;
    }

    public void close() {
    }

    public char read()
    {
        char ch;
        try
        {
            if (used == typed)
            {
                typed = 0;
                used = 0;
                ch = (char) in.read();
                while (ch != IO.EOF && ch != IO.CR && ch != IO.NL)
                {
                    buffer[typed] = ch;
                    typed++;
                    ch = (char) in.read();
                }
                if (IO.OS == IO.OS_WINDOWS && ch == IO.CR)  {
                    ch = (char) in.read();
                }
                String s = (new String(buffer)).substring(0,typed);
                if (s.equals(IO.EOF_STRING)) {
                    buffer[0] = IO.EOF;
                    typed = 1;
                    used = 0;
                }
                else if (ch == IO.EOF) {
                    buffer[typed] = IO.EOF;
                    typed++;
                }
                else {
                    buffer[typed] = IO.NL;
                    typed++;
                }
            }
            ch = buffer[used];
            if (ch != IO.EOF)  {
                used = used + 1;
            }
            return ch;
        } catch (Exception e)
        {
            IO.error("problems reading System.in");
            return IO.EOF;
        }
    }
}

/* THE INFILE CLASS
===========================================================
Implements a Reader interface for inputting single
characters only from a disk file.
 */

class Infile implements Reader
{
    private InputStream in;

    public Infile(String name) throws Exception
    {
        in = new FileInputStream(name);
    }

    public void close() throws Exception
    {
        if (in != null) {
            in.close(); in = null;
        }
    }

    public char read()
    {
        try {
            char ch = (char)in.read();
            switch (IO.OS) {
                case IO.OS_MAC:
                    if (ch == IO.CR) ch = IO.NL;
                    break;
                case IO.OS_UNIX:
                    /*skip*/ break;
                case IO.OS_WINDOWS:
                    while (ch == IO.CR) {
                        ch = (char)in.read();
                    }
                    break;
            }
            return ch;
        } catch (Exception e) {
            IO.error("problems reading from input file");
            return IO.EOF;
        }
    }
}

/* THE INPUT CLASS
===========================================================
Used to input a text file from a Keyboard or disk.
 */

class Input
{
    private Reader in;
    private boolean ahead;
    private char ch /*next character (if ahead)*/;

    public boolean done; /*true if previous input operation was successful*/

    public Input()
    {
        try {
            in = new Keyboard("ISO-8859-1");
            done = true;
        } catch (Exception e) {
            IO.error("could not open Keyboard");
            done = false;
        }
        ahead = false;
    }

    public Input(String name)
    {
        try {
            in = new Infile(name);
            done = true;
        } catch (Exception e) {
            IO.error("could not open input file " + name);
            done = false;
        }
        ahead = false;
    }

    public Input(Reader rdr)
    {
        in = rdr;
        done = true;
        ahead = false;
    }

    public boolean blank()
    {
        getAhead();
        return (ch == IO.CR) | (ch == IO.NL) | (ch == IO.SPACE);
    }

    public void close()
    {
        try {
            if (in != null) {
                in.close();
                in = null;
            }
            done = true;
        } catch (Exception e) {
            IO.error("could not close input");
            done = false;
        }
    }

    public boolean digit()
    {
        getAhead();
        return ('0' <= ch) & (ch <= '9');
    }

    private void getAhead()
    {
        if (!ahead) {
            readNext();
        } else {
           done = ch != IO.EOF;
        }
    }

    public boolean letter()
    {
        getAhead();
        return  (('a' <= ch) & (ch <= 'z')) |
            (('A' <= ch) & (ch <= 'Z'));
    }

    public boolean more()
    {
        getAhead();
        return ch != IO.EOF;
    }

    public char next()
    {
        getAhead();
        return ch;
    }

    public char read()
    {
        getAhead();
        ahead = false;
        return ch;
    }

    public void readBlanks()
    {
        while (blank()) readNext();
    }

    public boolean readBoolean()
    {
        boolean value; String symbol = readName();
        if (symbol.equals("false")) {
            value = false;
        } else if (symbol.equals("true")) {
            value = true;
        } else {
            value = false;
            done = false;
        }
        return value;
    }

    public double readDouble()
    {
        StringBuffer symbol = new StringBuffer();
        readBlanks();
        if (ch == '+') {
            readNext();
        }
        else if (ch == '-') {
            symbol.append(read());
        }
        while (digit()) {
            symbol.append(read());
        }
        if (ch == '.') {
            symbol.append(read());
            while(digit()) {
                symbol.append(read());
            }
        }
        if ((ch == 'e') | (ch == 'E'))
        {
            symbol.append(read());
            symbol.append(readInt());
        }
        try {
            Double numeral = new Double(symbol.toString());
            done = true;
            return numeral.doubleValue();
        } catch (Exception e) {
            done = false;
            return 0;
        }
    }

    public int readInt()
    {
        StringBuffer symbol = new StringBuffer();
        readBlanks();
        if (ch == '+') {
            readNext();
        } else if (ch == '-') {
            symbol.append(read());
        }
        while (digit()) {
            symbol.append(read());
        }
        try {
            done = true;
            return Integer.parseInt(symbol.toString(), 10);
        } catch (Exception e) {
            done = false; return 0;
        }
    }

    public String readLine()
    {
        StringBuffer line = new StringBuffer();
        getAhead();
        while (ch != IO.NL && ch != IO.EOF) {
            line.append(ch);
            readNext();
        }
        ahead = false;
        done = done || line.length() > 0;
        return line.toString();
    }

    public void readLn()
    {
       readLine();
    }

    public String readName()
    {
        StringBuffer letters = new StringBuffer();
        readBlanks();
        while (letter() | digit() | (ch == '_'))
        {
            letters.append(ch); readNext();
        }
        done = letters.length() > 0;
        return letters.toString();
    }

    public void readNext() {
        ch = (char)in.read();
        ahead = true;
        done = ch != IO.EOF;
    }

    public String readWord()
    {
        StringBuffer word = new StringBuffer();
        readBlanks();
        while (!blank())
        {
            word.append(ch); readNext();
        }
        done = word.length() > 0;
        return word.toString();
    }
}

/* THE RANDOM CLASS
============================================================
Used to generate random integers (or doubles) which are
uniformly distributed in a given range.
 */

class Random
{
    private static final double
    a = 16807.0,
    b = 2147483647.0;
    private double seed;
    private double m, n;

    public Random(int min, int max)
    {
        IO.assume(min <= max, "invalid random bounds");
        long time = System.currentTimeMillis();
        seed = (double)(time%1000000 + 1);
        m = (double)min; n = (double)max;
    }

    private double f()
    /* 0.0 <= f < 1.0 */
    {
        double temp = a*seed;
        seed = temp - b*Math.floor(temp/b);
        return seed/b;
    }

    public double readDouble()
    /* m <= readDouble < n */
    {
        return m + f()*(n - m);
    }

    public int readInt()
    /* m <= readInt <= n */
    {
        return (int)Math.floor(m + f()*(n - m + 1));
    }

    public void setSeed(int n)
    {
        seed = (double)(Math.abs(n)%1000000 + 1);
    }
}
